import re


# ------------------------------------------------------------------------------
class RePattern(object):

    def __init__(self, pattern, b_pattern=None):
        self.pattern = pattern
        self.b_pattern = b_pattern

    def __str__(self):
        return '{name}("{pattern}")'.format(
            name=type(self).__name__,
            pattern=self.pattern.pattern,
        )

    def search(self, text):
        return self.pattern.search(text)

    def b_search(self, text):
        return self.b_pattern.search(text)


# ------------------------------------------------------------------------------
class Command(RePattern):

    def __init__(self, cmd, flags=None, binary=True):
        self.cmd = cmd
        self.flags = flags

        pattern = self.compile_pattern(binary=False)
        b_pattern = self.compile_pattern(binary=True) if binary else None

        super(Command, self).__init__(pattern, b_pattern=b_pattern)

    def __str__(self):
        return '{name}("{command}", flags={flags})'.format(
            name=type(self).__name__,
            command=self.cmd,
            flags=', '.join('({}/{})'.format(*short_long) for short_long in self.flags or []),
        )
    
    def compile_pattern(self, binary=False):
        cmd = self.cmd

        if self.flags:
            all_short_long_flags = '|'.join('(?:{}|{})'.format(*short_long) for short_long in self.flags)
            cmd += r'\s[^;]*-' + all_short_long_flags
        else:
            cmd += r'(?:$|\s+\w+|(?:\s*[\\;\(\-\"]+(?!=)))'

        pattern = (
            r'(?:^|(?<=[`;\"\(])|(?<=\\n))\s*' +  # Command either starts a line or begins with ;(`"
            cmd
        )
        if binary:
            pattern = pattern.encode('utf-8')

        return re.compile(pattern)


UNSAFE_COMMANDS = [
    Command('eval'),
    Command('evalDeferred'),
    Command('evalEcho'),
    Command('evalNoSelectNotify'),
    Command('exec'),
    Command('executeForEachObject'),
    Command('python'),
    Command('callPython'),
    Command('system'),
    Command('popen'),
    Command('loadModule'),
    Command('source'),
    Command('pluginInfo'),
    Command('loadPlugin'),
    Command('unloadPlugin'),
    Command('catch'),
    Command('catchQuiet'),
]

# -- Overrides
UNSAFE_COMMANDS += [
    Command('alias'),
    Command('assignCommand'),
    Command('nameCommand'),
    Command('runTimeCommand'),
    Command('proc'),
    Command('global proc'),
]

# -- Callbacks
UNSAFE_COMMANDS += [
    Command('callbacks'),
    Command('file', [('c', 'command')]),
    Command('scriptJob'),
]

# -- Nodes
UNSAFE_COMMANDS += [
    Command('scriptNode'),
]

# -- Web
UNSAFE_COMMANDS += [
    Command('launch'),
    Command('putenv'),
]

# -- Ui elements
UNSAFE_COMMANDS += [
    Command('headsUpDisplay'),
    Command('hotBox', flags=[('ncc', 'noClickCommand')]),
    Command('window'),  # In attempt to catch attempts to create new ui elements with callbacks
    Command('menu'),
    Command('outlinerEditor', flags=[('sec', 'selectCommand')]),
    Command('modelEditor', flags=[('ec', 'editorChanged')]),
]

# -- Misc
UNSAFE_COMMANDS += [
    Command('fwrite'),
    Command('fwriteAllLines'),
    Command('fwriteAllText'),
    Command('fprint'),
    Command('cmdFileOutput'),
    Command('scriptEditorInfo', flags=[('wh', 'writeHistory')]),
]


ALL_COMMANDS = re.compile('|'.join(pat.pattern.pattern for pat in UNSAFE_COMMANDS), flags=re.DOTALL)


def scan_statement_commands(statement):
    # Also scan the setAttr string values.
    if 'setAttr ' == statement[:8]:
        try:
            command, attr, type_flag, type_value, value = statement.split(None, 4)

            # Skip non string/stringArray setAttr types
            if (type_flag != '-type' or
                    type_value not in ('"string"', '"stringArray"')):
                return None

            if ALL_COMMANDS.search(value):
                for command in UNSAFE_COMMANDS:
                    if command.search(value):
                        return '"{}" found in setAttr value!'.format(command), statement

        except ValueError:
            pass

    if ALL_COMMANDS.search(statement):
        for command in UNSAFE_COMMANDS:
            if command.search(statement):
                return '"{}" found!'.format(command), statement

    return None


# ------------------------------------------------------------------------------
# As we're checking to match any characters that can be used to
# call additional commands. This is difficult to spot with huge strings.
# setAttr ".b" -type "string" ("This is ");python("import os");$v=(" not the same string")
#         |__|       |______|  |________| ^       |_________| ^    |____________________|
MODIFIED_QUOTE_CAPTURE_PATTERN = re.compile(
    r'(?!\\)\"(?:[^\"\\]|\\.)*\"',  # Capture anything between unescaped quotes.
    flags=re.MULTILINE | re.DOTALL,
)
MODIFIED_PROHIBITED_ROOT_CHARACTERS = re.compile('[;`]')
MODIFIED_INLINE_FUNCTION_PATTERN = re.compile(r'\b(?<!-|:|\|)[a-zA-Z]+')


def scan_statement_modified(statement):
    end = 0
    for quote_match in MODIFIED_QUOTE_CAPTURE_PATTERN.finditer(statement):
        start = quote_match.start()

        # Check for any ; between captured quotes as its a convenient place to hide.
        between_quote = statement[end:start]
        for prohibited_match in MODIFIED_PROHIBITED_ROOT_CHARACTERS.finditer(between_quote):
            return '"{}" found - signs of file-tampering!'.format(prohibited_match.group()), statement

        if end > 0:
            for inline_match in MODIFIED_INLINE_FUNCTION_PATTERN.finditer(between_quote):
                return '"{}" found between quotes - signs of file-tampering!'.format(inline_match.group()), statement

        end = quote_match.end()

    return None


# ------------------------------------------------------------------------------
PLAINTEXT_PATTERNS = (
    'createNode script -n "vaccine_gene"',
    'createNode script -n "breed_gene"',
    '$fuck_All_U',
)


def scan_statement_plaintext(statement):
    for pattern in PLAINTEXT_PATTERNS:
        if pattern in statement:
            return '"{}" found - known viruses!'.format(pattern), statement

    return None
