import re


ASCII_STANZA_PATTERN = re.compile(
    '^'
    r'(?!//.+?\n)'  # Ignore comments
    r'(.+?(?:;|\n))(?:\n|$)',  # Get all characters up until ";\n" or a next valid statement.
    flags=re.MULTILINE | re.DOTALL,
)

INLINE_COMMAND_PATTERN = re.compile(
    r'`[\w\s\.\$\-(?:\"\")]+`',
    flags=re.MULTILINE | re.DOTALL,
)

_SAFE_STATEMENT_PATTERNS = (
    re.compile(
        r'^createNode '
        r'(?!(?:script|expression))'  # Dont match script or expression
        r'\w+ '  # Node type (no quotes, alphanumerical)
        r'(-s )?'  # Optional flag
        r'-n \"[\w\|\:]+\"'  # Known flags and alphanumerical long namespaced name
        r'(\s+-p\s+\"[\w\|\:]+\")?'  # Optional parent and alphanumerical long namespaced name
        r'$'
    ),
)


# ------------------------------------------------------------------------------
def _filter_matches(match):
    # Strip whitespace and the trailing ;
    # This reduces the complexity of the scanning.
    stanza = match.group().strip()[:-1]

    # Just to be safe, give any statements containing
    # backticks or basic concatenation operators.
    if ('`' in stanza or
            '(' in stanza or
            ';' in stanza or
            '{' in stanza):
        return match

    elif 'fileInfo ' == stanza[:9]:
        return None

    elif 'addAttr ' == stanza[:8]:
        return None

    elif 'rename -uid ' == stanza[:12]:
        return None

    elif 'select -ne :' == stanza[:12]:
        return None

    elif 'connectAttr ' == stanza[:12]:
        return None

    # Return only setAttrs that are string or stringArray.
    # Whilst malicious data could be stored in other formats, its useless
    # without a way to execute it.
    elif 'setAttr ' == stanza[:8]:
        try:
            command, value, typ, type_value = stanza.split(None, 4)
            if typ == '-type' and type_value in ('"string"', '"stringArray"'):
                return match
            else:
                return None

        except ValueError:
            return None

    for safe_pattern in _SAFE_STATEMENT_PATTERNS:
        if safe_pattern.match(stanza):
            return None

    return match


def _preprocess_stream(stream):
    all_matches = ASCII_STANZA_PATTERN.finditer(stream.read())
    filtered_matches = filter(None, (_filter_matches(match) for match in all_matches))
    return filtered_matches


def parse_stream(stream):
    statements = _preprocess_stream(stream)
    flatten_str_add_ops_pattern = re.compile(r'\"\s*\+\s*\"')
    for match in statements:
        stanza = match.group().strip()[:-1]
        if '+' in stanza:
            stanza = flatten_str_add_ops_pattern.sub('', stanza)

        yield stanza


def parse_file(self, filepath):
    with open(filepath, 'rt') as fp:
        for stanza in self.parse_stream(fp):
            yield stanza
