"""
manik_plugin.py

This is a maya plugin wrapper for the manik security scanner.  The primary
use is to enable a callback which scans MA files for potential security
problems on open.

The plugin also exposes a mel command `manik` which
takes a file or folder path argument.  The manik command
will scan the contents of a file or folder looking for
potential MA security issues.  Results will be printed to the listener

IMPORTANT:

this is a BETA version of manik plugin from early 2021
be sure to check for updates on a regular basis!
"""

import sys
import site
import os
import inspect
import maya.api.OpenMaya as om
import maya.cmds as cmds

filename = inspect.getframeinfo(inspect.currentframe()).filename
path = os.path.dirname(os.path.abspath(filename))
site.addsitedir(path + "/manik.zip")

import manik.runtime
import manik.scan
import manik.semantic.api


__version__ = (0, 1, 4)


def maya_useNewAPI():
    """This is an api plugin"""
    pass


# command
class ManikCommand(om.MPxCommand):

    kPluginCmdName = "manik"

    def __init__(self):
        om.MPxCommand.__init__(self)

    @staticmethod
    def cmdCreator():
        return ManikCommand()

    def doIt(self, args):
        """
        the 'manik' command runs a file or folder scan.  Running this inside of
        Maya is slower than running it standalone -- see the manik docs
        for how to run a bulk scan from the commandline
        """
        syntax = om.MSyntax()
        syntax.addArg(om.MSyntax.kString)
        argData = om.MArgParser(syntax, args )
        file_or_folder = (argData.commandArgumentString(0))
        if not os.path.exists(file_or_folder):
            raise RuntimeError("path {} does not exist".format(file_or_folder))

        scanner = manik.scan.MAScanner.from_module(manik.semantic.api)

        cmds.warning("scanning {}".format(file_or_folder))
        cmds.warning("this can be quite slow on large folders or files, consider running in a standalone")

        results = []
        try:
            cmds.waitCursor(st=1)
            if os.path.isdir(file_or_folder):
                # scan directory
                folder_list = [file_or_folder]
                for root, folders, _ in os.walk(file_or_folder):
                    for f in folders:
                        folder_list.append(os.path.join(root, f))

                # Give the results in some meaningful way.
                # Return only files with issues with the filepath followed by the result.
                # Mimics other maya commands (for better or for worse) so shouldn't
                # feel "out-of-place".
                scan_result = scanner.scan_folders(*folder_list)
                for filepath, file_scan_result in scan_result.items():
                    if file_scan_result:
                        results.append(filepath)
                        results.append(str(file_scan_result))

            else:
                scan_result = scanner.scan_file(file_or_folder)
                results = [file_or_folder, str(scan_result)] if scan_result else []

        finally:
            cmds.waitCursor(st=0)

        # note this returns a single-item array -- that's not by design,
        # its a limitation of the api plugin
        self.setResult(results)


class ManikCallbackCommand(om.MPxCommand):

    kPluginCmdName = "manik_callback"

    def __init__(self):
        om.MPxCommand.__init__(self)

    @staticmethod
    def cmdCreator():
        return ManikCallbackCommand()

    def doIt(self, arg):
        """
        enable beforeOpen callback protection
        """
        syntax = om.MSyntax()
        syntax.addArg(om.MSyntax.kBoolean)
        argData = om.MArgParser(syntax, arg)
        val = argData.commandArgumentBool(0)
        if val:
            manik.runtime.register_callbacks()
            sys.stdout.write("Manik protection enabled")
        else:
            manik.runtime.unregister_callbacks()
            sys.stdout.write("Manik protection disabled")


def initializePlugin(plugin):

    pluginFn = om.MFnPlugin(plugin)
    try:
        pluginFn.registerCommand(
            ManikCommand.kPluginCmdName,
            ManikCommand.cmdCreator
        )
    except:
        sys.stderr.write(
            "Failed to register command: %s\n" % ManikCommand.kPluginCmdName
        )
        raise

    try:
        pluginFn.registerCommand(
            ManikCallbackCommand.kPluginCmdName, 
            ManikCallbackCommand.cmdCreator
        )
    except:
        sys.stderr.write(
            "Failed to register command: %s\n" % ManikCallbackCommand.kPluginCmdName
        )
        raise

    manik.runtime.register_callbacks()
    cmds.warning("Manik protection enabled")


def uninitializePlugin(plugin):

    pluginFn = om.MFnPlugin(plugin)
    try:
        pluginFn.deregisterCommand(ManikCommand.kPluginCmdName)
    except:
        sys.stderr.write(
            "Failed to unregister command: %s\n" % ManikCommand.kPluginCmdName
        )
        raise

    try:
        pluginFn.deregisterCommand(ManikCallbackCommand.kPluginCmdName)
    except:
        sys.stderr.write(
            "Failed to unregister command: %s\n" % ManikCallbackCommand.kPluginCmdName
        )
        raise

    manik.runtime.unregister_callbacks()
    cmds.warning("Manik protection disabled")
