from __future__ import print_function
import pyparsing as pp
from pyparsing import (
    Combine,
    dblQuotedString,
    delimitedList,
    Forward,
    Group,
    Keyword,
    LineEnd,
    Literal,
    nestedExpr,
    NotAny,
    oneOf,
    OneOrMore,
    Optional,
    PrecededBy,
    QuotedString,
    Suppress,
    Word,
    WordStart,
    ZeroOrMore,
) 

"""
This file defines pyparsing constructs for the basic elements of MEL syntax. 

It' not a 'real' MEL parser because it doesn't attempt to determine the validity of the
underlying MEL -- it only knows enough to get the basics of what MEL expressions and statements
look like so it can efficiently deconstruct an MA file looking for potentially evil 
(but syntactically correct) statements.

Note that the way pyparsing aggregates these checks order is usually very important! Be careful
if you move expressions around inside a parsing expression

The main public interfaces of the file is the method get_parser, which will
return a pyparsing object capable of decyphering all of the mel expressions in a typical MA file,
including the mel scripts escaped into string attributes

See the pyparsing docs ( https://pyparsing-docs.readthedocs.io/en/latest/HowToUsePyparsing.html )
for more detail on the returned objects.
"""


common = pp.pyparsing_common

MEL_IDENTIFIERS = pp.alphas + "_01234567890"
BASE_64 = pp.alphas + "+-/_"
BINARY = "\\x0123456789abcdefgh"


"""
These keywords can be extracted from a PyParsing.ParseResult to find important
elements in a mel statement:  for example

      result.get(KEYWORD_TOKEN)

will return the command from a mel statement, if there is one, or None
"""

# a command, typically but not always the first word in a statement
KEYWORD_TOKEN = "m_keyword"

# a string literal -- usually a string attribute value but also a name
STRING_TOKEN = "m_str"

# a mel call enclosed in backticks
CALL_TOKEN = "m_call"

# a c-style mel function call, ie "someFunction()"
FUNC_TOKEN = "m_func"

# a mel imperative command with flags and arguments; most of the file
COMMAND_TOKEN = "m_imp"

# a mel procedure declaration. if this appears in the wild it's a bad sign
PROC_TOKEN = "m_proc"

# a flag inside a mel command
FLAG_TOKEN = "m_flag"

# an argument list inside a mel command
ARGS_TOKEN = "m_arg"

CONDITIONAL_TOKEN = "m_conditional"

"""
Below here we assemble a rudimentary mel grammar from pyParsing expressions
"""


# line terminato: {expr}r
semicolon = Literal(";").suppress()

# number values (using pp.common to get negatives, sci notation, etc)
mel_number = (
    common.sci_real | common.real | common.signed_integer | pp.nums
).setResultsName("m_num")


# mel data types
mel_types = oneOf(("string", "float", "vector", "matrix", "int"), asKeyword=True)
mel_flowcontrol = oneOf(
    (
        "if",
        "then",
        "else",
        "try",
        "catch",
        "catchQuiet",
        "break",
        "continue",
        "do",
        "while",
        "for",
        "in",
    ),
    asKeyword=True,
)
mel_directives = oneOf(("global", "proc", "env", "rehash", "source"))

mel_bools = oneOf("true false")

# all reserved words
reserved = mel_flowcontrol | mel_types | mel_directives | mel_bools

# generic mel tokens (we only expect  one of these at the root of a statement)
command = (
    NotAny(reserved) + WordStart() + Word(pp.alphas, MEL_IDENTIFIERS)
).setResultsName(KEYWORD_TOKEN)

# Mel allows unquoted object names, and those sometimes incldue namespaces.
mel_constant = (
    NotAny(reserved) + Word(":" + MEL_IDENTIFIERS) | Word("%" + MEL_IDENTIFIERS)
).setResultsName(
    "m_constant"
)  # eg, an unquoted namespace


# quoted string literals (mel only uses double quotes)
# this does not handle multi-line strings : those are treated as binary additions
mel_string_literal = dblQuotedString().setResultsName(STRING_TOKEN)
# POSSIBLE ISSUE: not sure how to handle escapes...


# vector and matrix literals
number_list = delimitedList(mel_number, ",")
mel_matrix_literal = (
    Suppress("<<") + delimitedList(number_list, ";") + Suppress(">>")
).setResultsName("m_matrix")
mel_vector = Suppress("<<") + number_list + Suppress(">>").setResultsName("m_vec")
mel_multi = Group(mel_matrix_literal | mel_vector).setResultsName("m_array")

# Array literals
mel_array_literal = Group(
    Suppress("{")
    + delimitedList(mel_vector | mel_number | mel_string_literal, ",")
    + Suppress("}")
).setResultsName("m_array_literal")

# mel variables (with optional subscripts)
mel_subscript = (Literal("[") + Optional(Word(pp.nums)) + Literal("]")).setResultsName(
    "m_subscript"
)
mel_var = Word("$", MEL_IDENTIFIERS, min=2).setResultsName("m_varname")
mel_variable = Combine(
    mel_var + Optional(mel_subscript) + Optional(mel_subscript)
)._setResultsName("m_var")


# mel call using backticks.  We  don't really expect to see this at the file level,
# so it's treated as a string since it's almost certainly trouble.  The contents
# are interpreted as mel by the parse action
mel_backtick = QuotedString("`", "`")  # .setResultsName(CALL_TOKEN)


def parse_backtick(s, lok, toks):
    result = valid_mel_line.setResultsName(CALL_TOKEN).parseString(" ".join(toks))
    return result


mel_backtick.setParseAction(parse_backtick)

# an operand is something that can be part of an operation:
# a constant, a variable, a literal, a backticked call, or a function call,
mel_operand = (
    mel_number
    | mel_constant
    | mel_variable
    | mel_vector
    | mel_multi
    | mel_string_literal
    | mel_array_literal
    | mel_backtick
)


# operators
# note we want the single character ones to come _late_
mel_op = oneOf("|| && += -= /= *= == <= >= = > < + - / * % ^", asKeyword=True)

# 'Forward is how pyparsing handles recursive references
# note we need to define func calls inside of the binary op too -- a func call
# can include an operators like `call($a + $b)` so theres double
# recursion here
mel_binary_op = Forward()
mel_func_call = Forward()
func_argument = mel_binary_op | mel_operand | mel_func_call
argument_list = Optional(delimitedList(func_argument))
mel_func_call << (
    NotAny(reserved)
    + Word(pp.alphas, MEL_IDENTIFIERS, asKeyword=True).setResultsName(KEYWORD_TOKEN)
    + Group(Suppress("(") + argument_list + Suppress(")")).setResultsName("m_fun_args")
).setResultsName(FUNC_TOKEN)


parenthesized = Group(Suppress("(") + mel_binary_op + Suppress(")"))
mel_atom = (mel_func_call | parenthesized | mel_operand).setResultsName("m_atom")
mel_binary_op << Group(mel_atom + ZeroOrMore(mel_op + mel_atom, stopOn=";"))
mel_binary_op.setDefaultWhitespaceChars(" \t")  # can be multiline
mel_binary_op.setResultsName("m_binop")

# unary ops (mel does not handle (-$a, so this is just ++ and --))
mel_unary_op = Combine(mel_var + (Literal("++") | Literal("--"))).setResultsName(
    "m_unop"
)

# mel supports ternary ops   (condition) ? val1 : vale2
mel_ternary_op = (
    (mel_atom | parenthesized)
    + Literal("?")
    + (mel_binary_op | mel_atom)
    + Literal(":")
    + (mel_binary_op | mel_atom)
).setResultsName("m_ternary")

# classic mel-style command formatting   command "arg"... "argN" -flag 1 -otherFlag
mel_flag = Word("-", pp.alphas, min=2, asKeyword=True).setResultsName("m_flag_name")
allowable_values = (
    NotAny(semicolon) + NotAny(mel_flag) + (mel_func_call | mel_atom | mel_binary_op)
).setResultsName("m_flag_val")
flag = Group(mel_flag + ZeroOrMore(allowable_values) + Optional(Suppress(LineEnd())))


args = Group(ZeroOrMore(allowable_values)).setResultsName(ARGS_TOKEN)
imperative = (command + Optional(args) + ZeroOrMore(flag)).setResultsName(COMMAND_TOKEN)


# a mel comment (//....) a pyparsing builtin
mel_comment = (pp.cppStyleComment()).setResultsName("m_comment")

# return statements
mel_return = Group(Word("return", asKeyword=True) + (mel_binary_op | mel_atom))


valid_mel_line = Forward()
# conditionals.  These should only ever occur within a script,
# which makes them inherently suspect.  For now we capture the
# associated blocks verbatim since any code like this in-line
# is almost certainly unwanted
mel_block = Group(
    Literal("{").suppress()
    + Optional(OneOrMore(LineEnd()).suppress())
    + OneOrMore(valid_mel_line).setResultsName("contents")
    + Literal("}").suppress()
    + Optional(Suppress(LineEnd()))
).setResultsName("m_block")

# this implementation does not consider the contents of the
# conditional expression, maybe it should?
mel_conditional = nestedExpr("(", ")").setResultsName("m_cond")
else_block = Keyword("else") + Optional(LineEnd()) + mel_block
mel_if = (
    Keyword("if")
    + (mel_conditional | mel_atom)
    + Optional(LineEnd())
    + mel_block.setResultsName("block_1")
    + Optional(else_block).setResultsName("block_2")
)

mel_while = (
    Keyword("while")
    + (mel_conditional | mel_atom)
    + (mel_block.setResultsName("block_1"))
)
mel_do = (
    Keyword("do")
    + (mel_block.setResultsName("block_1"))
    + Keyword("while")
    + mel_conditional
)
mel_for = (
    Keyword("for")
    + mel_conditional
    + Optional(LineEnd())
    + (mel_block.setResultsName("block_1"))
)
mel_escape = oneOf("break continue")
mel_flow = (
    (mel_if | mel_for | mel_while | mel_for | mel_escape) + Optional(LineEnd())
).setResultsName(CONDITIONAL_TOKEN)

# mel_style function   proc returnType name (int $arg1, float $arg2)  etc
arg_spec = mel_types + mel_var
cmd_args = Group(
    Suppress("(") + Optional(delimitedList(arg_spec, ",")) + Suppress(")")
).setResultsName("cmd_args")

mel_procedure = (
    Optional(Word("global"))
    + Word("proc", asKeyword=True)
    + Optional(mel_types)
    + Word(pp.alphanums).setResultsName("m_proc")
    + cmd_args
    + Optional(LineEnd())
    + Optional(mel_block)
).setResultsName(PROC_TOKEN)
# note the mel block is "optional" in case the  the preprocessor
# or a hacker breaks the stanza before the proc body. We
# probably only care about the existence of a definitions


# assignments
mel_assign = Group(
    Optional(Word("global", asKeyword=True))
    + Optional(mel_types)
    + mel_variable
    + Literal("=")
    + mel_atom
).setResultsName("m_assign")

# declarations
mel_declare = Group(
    Optional(Word("global", asKeyword=True)) + Optional(mel_types) + mel_variable
).setResultsName("m_declare")

# mel statements:  a complete mel instruction, either a line of a script
mel_stmt = (mel_return | mel_func_call | mel_unary_op | mel_ternary_op).setResultsName(
    "m_stmt"
)

# all the possible outcomes in a single block, ending in a semicolon. This is the
# key parsing component used to parse a file

valid_mel_line <<= (
    (
        mel_comment
        | mel_stmt
        | mel_flow
        | mel_procedure
        | imperative
        | (mel_assign | mel_declare)
    )
    + Optional(semicolon)
    + Optional(LineEnd().suppress())
)


def get_parser():
    """
    this is the public interface to the file: it returns as pyParsing.ParseExpression
    node which knows how to iterative decompose a string or stringstream inton
    tokens.  Example use:

        def parse_stream(filehandle):

            text = filehandle.read()
            for match, start, stop in valid_mel_line.scanString(text):
                yield match, start, stop

    Note that in practice you want to use a preprocessor to keep the incoming texts short --
    that makes the parser run a lot more efficiently.
    """
    output_parser = valid_mel_line.copy().setResultsName("mel_stanza")
    output_parser.setDefaultWhitespaceChars(" \t")
    return output_parser


def get_binary_parser():
    """
    a separate pass on a chunk that identifies binary data independent of structure parsing
    """
    b64data = Word(BASE_64, min=32).setResultsName("base64")
    binary = Word(BINARY, min=64).setResultsName("binary")
    return pp.OneOrMore(b64data | binary)
