from __future__ import print_function, division
import logging
import sys
import ast
import re


logger = logging.getLogger("manik.semantic.preprocessor")

PY2 = sys.version_info[0] < 3

"""
This module contains helpers to be run before doing semantic analysis on the file contents.
"""


# these mark potentially executable code extracted from
# string attributes.  They originate here becuase this module

SUSPECT_STRING = "suspect_string"
CALLABLE_STRING = "callable_string"


def str_split(text, separator):
    """
    Efficient str.split generator alternative.
    https://stackoverflow.com/a/65617215
    """
    i = 0
    n = len(text)
    while i <= n:
        j = text.find(separator, i)
        if j == -1:
            j = n
        yield text[i:j]
        i = j + 1


def yield_clauses(stream):
    """
    Loop through the  file-like object <stream>, yielding out semicolon-delimited statements.  Semicolons embedded in quotes
    (particularly those long uiConfiguration script nodes, but any text field containing mel) will be ignored.  Preprocessing the file
    like this helps simplify the parsing stream
    """

    cache = stream.read()
    line_count = 0

    # Using a split generator avoids excessive memory consumption on large
    # scenes with seemingly little performance cost compared to `str.split()`.
    chunks = str_split(cache, ";\n")
    # chunks = cache.split(";\n")

    for c in chunks:
        line_count += c.count('\n') + 1
        potentially_inlined = "`" in c or "(" in c
        yield c.strip(), line_count, potentially_inlined


_flatten_str_add_ops_pattern = re.compile(r'\"\s*\+\s*\"')


def unescape_text(textattr):
    """
    returns the contents of a possibly escaped
    string attribute as an unescaped string.
    """
    
    try:
        # the default case: it's just a string
        r = ast.literal_eval(textattr.strip())
        return r
    except:
        chunks = textattr.split("+")
        merged = "".join(chunks)
        try:
            # this should catch both MEL style string
            # concats (which are only legit for long strings)
            # if they are mel executable they will jump down to
            # one of the excsptions
            r = ast.literal_eval(merged)
            return r
        except SyntaxError:

            # this probably means something like a string attribute that
            # includes a mel backtick command that it's trying to execute
            # under the guise of string assembly, like
            # `eval "some attack code here"`
            return "{}({})".format(SUSPECT_STRING, textattr)
        except ValueError:
            # this means a an inline command using function syxtax
            # like eval("some attack code here)
            return "{}({})".format(CALLABLE_STRING, textattr)

def ignore_safe_calls(stanza_and_lineno_and_inlines):
    """
    filter out calls that are only dangerous if mangled, primarily 
    rename, connectAttr, relationship,
    """
    stanza, lineno, inlines = stanza_and_lineno_and_inlines
    if not inlines:
        cmd_names = (
            'rename',
            'connect',
            'relation',
            'parent',
            'select',
            'connectAttr'
        )
        token = stanza[:8]
        if any(cmd_name in token for cmd_name in cmd_names):
            logger.debug("---- %i %s...", lineno, stanza[:36])
            return "", lineno, False

    return stanza_and_lineno_and_inlines


def fork_setattr(stanza_and_lineno_and_inlines):
    """
    filters out setAttr calls which do not set string or stringArray attributes. This
    cuts out a lot of long, expensive calls that are not syntactically dangerous.

    Strictly speaking these calls might still hide bad data in all sorts of ways -- but
    they are ultimately harmless without some kind of string. This speeds up large model
    file processing by 90% or more.
    """

    stanza, lineno, inlines = stanza_and_lineno_and_inlines

    # pass non-setAttr through.. it's probably here
    # because it's got a suspect construct inline
    if stanza[:7] != "setAttr":
        return stanza, "", lineno

    # note long forms of these indicate a hand-written file,
    # so almost certainly malicious
    tokens = stanza[:64].split()

    # string attrib that might be dangerous
    if len(tokens) > 4 and '-type' in tokens:
        t = tokens.index("-type")
        # this logic handles legimitately formatted mel in strings,
        # like UI nodes
        if tokens[t+1] == '"string"':
            offset = len(" ".join(tokens[:t+2])) + 1
            return stanza, unescape_text(stanza[offset:].strip()), lineno

    # if we get here, the only triggers left are inline attacks
    if inlines:
        return stanza, "", lineno

    logger.debug("---- %i %s...", lineno, stanza[:36])
    return "", None, lineno


def prefilter_stream(stream):
    """
    Filters the file into a stream of easier-to-parse strings.

    The stream consists of tuples, the first item is the raw call
    (for example 'createNode') and the second item is the contents
    of any string attribute set by `setAttr` (for example the contents)
    of the uiConfiguration script nodes.  The second item is
    unescaped so other code does not need to worry about double quotes
    and so on.

    We separate these streams because executable code inline in the
    main file is always malicious where there is a lot of innocent
    code (such as the camera home commands) living in string attributes;
    forking them allows us to develop separate rules
    """
    stanzas_and_lines = yield_clauses(stream)
    remove_renames = (ignore_safe_calls(s) for s in stanzas_and_lines if s[0])
    forked = (fork_setattr(s) for s in remove_renames if s[0])
    dubious = (s for s in forked if s[0])

    return dubious
