"""
This file defines rules which will be run on file statements.
"""
try:
    from builtins import map
except:
    from itertools import imap as map
from collections import defaultdict
from datetime import date
import functools
import logging
import base64
from sre_constants import CHCODES
from sys import executable
import re
from unittest import runner

from manik.semantic.ma_preprocessor import SUSPECT_STRING, CALLABLE_STRING
from manik.semantic.mel_reference import (
    KNOWN_CALLBACK_ATTRIBS,
    MATH_COMMANDS,
    MEL_EVALS,
    MEL_FILE_COMMANDS,
    PYTHON_EXEC_COMMANDS,
    UI_COMMANDS,
    UI_CONF_BAD_FLAGS,
    UI_CONFIG_COMMANDS,
)


logger = logging.Logger("manik.semantic.rules")

_RULES = defaultdict(list)

def make_re_for_keyword(kw):
    return ("(" + kw + "\()")

executable_pattern = "|".join((k for k in MEL_EVALS)) + "|" + "|".join((k for k in PYTHON_EXEC_COMMANDS))  
executble_re = re.compile(executable_pattern)


def rule(key=None):
    """
    Decorate a single-argument callable with this function to register it as
    an analysis rule.  The optional key argument will register the rule in
    a category (for example, all scriptNode related filters could be registered
    using 'scriptNode')
    """

    def _deco(func):
        _RULES[key].append(func)

        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        return inner

    return _deco


"""
notes for writing rules:

rules process api.MelStatement objects, which are namedtuples (see the docs in api.py for details)

the 'context' field contains None if the statement is executed in the file body; if the
statement is extracted from a different statement, (for example, if this statement is the data
from a setAttr call) the context field contains the parent statement.  So, you can
test for file-level commands with 

    if stmt.context is None:

and for statements embedded in data with

    if stmt.context:

"""


@rule("scriptNode")
def scriptNode_cmd(stmt):
    """
    This statement contains a scriptNode command (not a createNode -type script)
    This means somebody is trying to trigger an attack
    """
    if stmt.command == "scriptNode":
        return "scriptnode command"


@rule("code")
def file_command(stmt):
    if stmt.command == "file" and ("-c" in stmt.flags or '-command' in stmt.flags):
        return "file command callback"


@rule("code")
def scriptEditorInfo_command(stmt):
    if stmt.command == "scriptEditorInfo" and ('-wh' in stmt.flags or '-writeHistory' in stmt.flags):
        return "scriptEditorInfo writeHistory command"


@rule("uiConfig")
def command_in_uiConfig(stmt):
    if stmt.context == "<UIConfig>":
        if stmt.command not in UI_CONFIG_COMMANDS:
            return "Illegal command in UIConfig"


@rule("code")
def is_eval(stmt):
    """
    variants of "eval", all gateways to arbitrary code
    """
    if stmt.is_definition:
        return

    if stmt.command in MEL_EVALS:
        return "eval statement"


@rule("code")
def system(stmt):
    if stmt.is_definition:
        return

    if stmt.command in ("exec", "system", "popen"):
        return "system call"


@rule("code")
def is_python(stmt):
    """
    Executable python
    """
    if stmt.is_definition:
        return

    if stmt.command in ("python", "pyCommand"):
        if stmt.context is None:
            return "python call in file body"
        elif stmt.context:
            return "python call in executable node"


@rule("code")
def is_alias(stmt):
    """
    an alias is probably somebody trying to get around a text scan for bad commands
    """
    if stmt.command == "alias":
        return "mel alias"


@rule("code")
def root_level_procedure(stmt):
    if stmt.is_definition and stmt.context is None:
        return "mel procedure definition in file body"


@rule("code")
def runtime_command(stmt):
    if stmt.command == "runTimeCommand":
        return "runtime Command"

@rule("code")
def callbacks_command(stmt):
    if stmt.command == "callbacks":
        return "callbacks command"


@rule("code")
def suspect_string(stmt):
    if stmt.command == SUSPECT_STRING:
        return "suspicious assembly in string attribute"


@rule("code")
def callable_string(stmt):
    if stmt.command == CALLABLE_STRING:
        return "callable code in string attribute"


@rule("scriptNode")
def scriptnode_creation(stmt):
    if stmt.command != "createNode":
        return

    if "script" in stmt.args:
        nodename = stmt.flags.get("-n", [""])[0]
        nodename = nodename[1:-1]  # remove the quotes
        nodename = nodename.rsplit(":", 1)[-1]
        if nodename not in (
            "sceneConfigurationScriptNode",
            "uiConfigurationScriptNode",
        ):
            return "file contains scriptNode"


@rule("code")
def python_fileopen(stmt):
    if stmt.command == "open" and len(stmt.args) == 2:
        return "Python file system access"


@rule("code")
def scriptJob(stmt):
    if stmt.command == "scriptJob":
        return "scriptJob command"


@rule("code")
def mel_file_commands(stmt):
    if stmt.is_definition:
        return

    if stmt.command in MEL_FILE_COMMANDS:
        return "file write operation"


@rule("code")
def mel_source(stmt):
    if stmt.command in ("source", "rehash"):
        return "MEL file source"

@rule("code")
def ui_command(stmt):
    if stmt.command in UI_COMMANDS:
        if not UI_CONF_BAD_FLAGS.get(stmt.command):

            if not stmt.context:
                return "UI Command in file body"
        else:
            allowed_flags = UI_CONF_BAD_FLAGS[stmt.command]
            for f in allowed_flags:
                if f in stmt.flags:
                    return "Illegal UI Callback flag"


@rule("data")
def executable_data(stmt):
    if stmt.command == "embedded":
        return "executable code in data statement"

@rule("data")
def embedded_in_data(stmt):
    if stmt.command == "embedded" and stmt.context is None:
        return "embedded"


@rule("data")
def gene(stmt):
    if stmt.command == "setAttr" and stmt.flags.get('-type'):
        if stmt.flags.get('-type')[0] == '"string"':
            value = stmt.flags.get('-type')[-1] 
            if isinstance (value, list):
                return "suspect string assembly"            


@rule("code")
def assignment(stmt):
    if not stmt.context and  stmt.command not in (
        'setAttr', 'addAttr', 'rename', 'requires', 'createNode', 'connectAttr', 'disconnectAttr', 'relationship', 'select', 'fileInfo', "file", "lockNode"
    ):
        return "mel command %s loose in file" % stmt.command

@rule("data")
def assigned_callbacks(stmt):

    if stmt.command == "setAttr":

        parent_attr = None
        for attr in stmt.args:
            if attr[:2] == '".':
                parent_attr = attr
                break

        if parent_attr == '".ixp"':
            if stmt.command not in MATH_COMMANDS:
                return "executable code in expression"

        elif parent_attr in ('".b"', '".before"'):
            # these are expected in the
            # ui config script nodes --
            # any other code is suspect
            if stmt.command not in UI_CONFIG_COMMANDS:
                return "unexpected code in UIConfig script"

        elif parent_attr in KNOWN_CALLBACK_ATTRIBS:
            return "executable code in callback"


def _isBase64(sb):
    try:
        if isinstance(sb, str):
            # If there's any unicode here, an exception will be thrown and the function will return false
            sb_bytes = bytes(sb, 'ascii')
        elif isinstance(sb, bytes):
            sb_bytes = sb
        else:
            raise ValueError("Argument must be string or bytes")
        return base64.b64encode(base64.b64decode(sb_bytes)) == sb_bytes
    except Exception:
        return False


@rule("data")
def binary_data(stmt):
    if stmt.context and stmt.command == "binary":
        if _isBase64(stmt.args[0]):
            return "binary data"


@rule("data")
def import_statement_in_data(stmt):
    """
    this is a good example of a rule that we might want to make
    contextual.
    """
    if stmt.is_definition:
        return

    for arg in stmt.args:
        if not isinstance(arg, str): 
            continue
        for c in PYTHON_EXEC_COMMANDS:
            if c in arg:
                return "{} statement in data".format(c)


@rule("ui")
def ui_callbacks(stmt):
    """
    if the attacker knows the name of any global Maya UI element they can easily
    update the callback of a button, menu item etc to trigger attack code when a user
    interacts with the UI.  Many of these also have executable callbacks.
    """
    if stmt.is_definition:
        return

    if stmt.command in UI_COMMANDS:
        stmt_flags = stmt.flags
        if "-q" in stmt_flags or "-query" in stmt_flags:
            return

        forbidden = UI_CONF_BAD_FLAGS.get(stmt.command)
        if forbidden is None:
            return #"UI Command"

        for f in forbidden:
            if f in stmt_flags:
                return "UI Command with harmful flag"


@rule("plugins")
def pluginInfo(stmt):
    if stmt.is_definition:
        return

    if stmt.command in ("pluginInfo", "loadPlugin", "unloadPlugin"):
        return "Plugin manipulation"


def get_rules(*categories):
    """
    returns a list of all the registered rules, or, if categories are supplied,
    return only the rules in those categories.
    """

    def list_join(a, b):
        a.extend(b)
        return a

    if categories:
        return functools.reduce(list_join, [_RULES.get(c) for c in categories])
    else:
        return functools.reduce(list_join, _RULES.values())
